from functools import partial
from typing import Any, Callable, Dict, List, Tuple

from wsgidav.mw.base_mw import BaseMiddleware  # type: ignore
from wsgidav.util import (  # type: ignore
    SubAppStartResponse,
    get_module_logger,
    init_logging,
)

_logger = get_module_logger(__name__)


class HeaderLogger(BaseMiddleware):
    def logger(
        self,
        environ: Dict[str, Any],
        start_response: Callable,
        status: int,
        headers: List[Tuple[str, str]],
        exc_info=None,
    ):
        _logger.debug(f"Request environ: {environ}")
        _logger.debug(f"Response headers: {dict(headers)}")
        return start_response(status, headers, exc_info)

    def __call__(
        self, environ: Dict[str, Any], start_response: Callable
    ) -> List[bytes]:
        return self.next_app(environ, partial(self.logger, environ, start_response))


def verbose_logging() -> None:
    # Enable everything that seems like module that could have logging
    init_logging(
        {
            "verbose": 5,
            "enable_loggers": [
                "manabi.log",
                "manabi.lock",
                "lock_storage",
                "lock_storage",
                "request_resolver",
                "request_server",
                "http_authenticator",
                "property_manager",
                "fs_dav_provider",
                "dir_browser",
                "server",
            ],
        }
    )


class ResponseLogger(BaseMiddleware):
    def __init__(self, application, next_app, config):
        self.__application = application
        self.next_app = next_app

    def __call__(self, environ, start_response):
        sub_app_start_response = SubAppStartResponse()
        first_yield = True
        app_iter = self.next_app(environ, sub_app_start_response)

        for v in app_iter:
            # Start response (the first time)
            if first_yield:
                # Success!
                start_response(
                    sub_app_start_response.status,
                    sub_app_start_response.response_headers,
                    sub_app_start_response.exc_info,
                )

            if environ["REQUEST_METHOD"] == "LOCK":
                _logger.debug(f"LOCK Response: {v.decode('utf-8')}")
            first_yield = False
            yield v

        if hasattr(app_iter, "close"):
            app_iter.close()

        # Start response (if it hasn't been done yet)
        if first_yield:
            # Success!
            start_response(
                sub_app_start_response.status,
                sub_app_start_response.response_headers,
                sub_app_start_response.exc_info,
            )

        return



import calendar
from datetime import datetime
from email.utils import formatdate
from http.cookies import SimpleCookie
from inspect import getsource
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import base62  # type: ignore
from attr import attrib, dataclass


def cattrib(
    attrib_type: Optional[type] = None,
    check: Optional[Callable] = None,
    optional: bool = False,
    **kwargs,
):
    if "default" in kwargs and kwargs["default"] is None:
        optional = True

    def handler(object, attribute, value):
        if value is None and optional:
            return value
        if attrib_type is not None and not isinstance(value, attrib_type):
            raise TypeError(
                f"{attribute.name} ({type(value)}) is not of type {attrib_type}"
            )
        if check and not check(value):
            source = getsource(check).strip()
            raise ValueError(f"check failed: {source}")
        return value

    return attrib(validator=handler, on_setattr=handler, **kwargs)


def get_rfc1123_time(secs: float = None) -> str:
    """Return <secs> in rfc 1123 date/time format (pass secs=None for current date)."""
    return formatdate(timeval=secs, localtime=False, usegmt=True)


def to_string(data: bytes) -> str:
    return base62.encodebytes(data)


def from_string(data: str) -> bytes:
    return base62.decodebytes(data)


@dataclass
class AppInfo:
    start_response: Callable = cattrib(check=lambda x: callable(x))
    environ: Dict[str, Any] = cattrib(dict)
    secure: bool = cattrib(bool, default=True)


def set_cookie(
    info: AppInfo,
    key: str,
    value: str,
    ttl: int,
    status: int,
    headers: List[Tuple[str, str]],
    exc_info=None,
):
    cookie: SimpleCookie = SimpleCookie()
    cookie[key] = value
    date = datetime.utcnow()
    unixtime = calendar.timegm(date.utctimetuple())
    cookie[key]["expires"] = get_rfc1123_time(unixtime + ttl)
    if info.secure:
        cookie[key]["secure"] = True
        cookie[key]["httponly"] = True
    headers.append(cast(Tuple[str, str], tuple(str(cookie).split(": "))))
    info.start_response(status, headers, exc_info)



from wsgidav import http_authenticator  # type: ignore
from wsgidav.http_authenticator import HTTPAuthenticator  # type: ignore

from .auth import ManabiAuthenticator


class MetaFakeHTTPAuthenticator(type(http_authenticator.HTTPAuthenticator)):  # type: ignore
    def __instancecheck__(cls, instance):
        return isinstance(instance, HTTPAuthenticator) or isinstance(
            instance, ManabiAuthenticator
        )


class FakeHTTPAuthenticator(
    http_authenticator.HTTPAuthenticator, metaclass=MetaFakeHTTPAuthenticator
):
    pass


# Instead of accepting an override for HTTPAuthenticator as for everything else, wsgidav
# just expects a class extending HTTPAuthenticator in the middleware_stack.
# wsgidav grabs that class out of the middleware_stack.
#
# Using a metaclass is slightly less intrusive than just replacing HTTPAuthenticator
# with ManabiAuthenticator. If http_authenticator.HTTPAuthenticator is created it is
# still a HTTPAuthenticator.
http_authenticator.HTTPAuthenticator = FakeHTTPAuthenticator


def keygen() -> None:
    from .util import to_string

    with open("/dev/random", "rb") as f:
        print(to_string(f.read(32)))



import fcntl
import shelve
from contextlib import contextmanager
from pathlib import Path

from wsgidav.lock_man.lock_storage import LockStorageDict  # type: ignore
from wsgidav.util import get_module_logger  # type: ignore

_logger = get_module_logger(__name__)


class ManabiLockLockStorage(LockStorageDict):
    def __init__(self, refresh: float, storage: Path):
        super().__init__()
        self.max_timeout = refresh / 2
        self._storage = storage

    @contextmanager
    def get_shelve(self):
        fd = self._lock_file.fileno()
        try:
            self._lock.acquire_write()
            try:
                if self._semaphore == 0:
                    fcntl.flock(fd, fcntl.LOCK_EX)
                    self._dict = shelve.open(str(self._storage))
                    _logger.debug(
                        f"get_shelve({self._storage}): {len(self._dict)} locks"
                    )
                self._semaphore += 1
                yield
            finally:
                self._semaphore -= 1
                if self._semaphore == 0:
                    self._dict.close()
                    self._dict = None
                    fcntl.flock(fd, fcntl.LOCK_UN)
        finally:
            self._lock.release()

    def open(self):
        _logger.debug(f"open({self._storage})")
        self._lock_file = open(f"{self._storage}.lock", "wb+")
        self._semaphore = 0

    def close(self):
        _logger.debug("close()")
        self._lock.acquire_write()
        try:
            lock = self._lock_file
            self._lock_file = None
            lock.close()
            self._semaphore = 0
        finally:
            self._lock.release()

    def clear(self):
        _logger.debug("clear()")
        self._lock.acquire_write()
        try:
            self._storage.unlink()
        finally:
            self._lock.release()

    def get(self, token):
        with self.get_shelve():
            return super().get(token)

    def create(self, path, lock):
        max_timeout = self.max_timeout
        timeout = lock.get("timeout")

        if not timeout:
            lock["timeout"] = max_timeout
        else:
            if timeout > max_timeout:
                lock["timeout"] = max_timeout
        with self.get_shelve():
            return super().create(path, lock)

    def refresh(self, token, *, timeout):
        with self.get_shelve():
            return super().refresh(token, timeout=timeout)

    def delete(self, token):
        with self.get_shelve():
            return super().delete(token)

    def get_lock_list(self, path, *, include_root, include_children, token_only):
        with self.get_shelve():
            return super().get_lock_list(
                path,
                include_root=include_root,
                include_children=include_children,
                token_only=token_only,
            )



from pathlib import Path
from typing import Any, Dict

from wsgidav.dav_error import HTTP_FORBIDDEN, DAVError  # type: ignore
from wsgidav.fs_dav_provider import (  # type: ignore
    FileResource,
    FilesystemProvider,
    FolderResource,
)

from .token import Token


class ManabiFolderResource(FolderResource):
    def get_member_names(self):
        token: Token = self.environ["manabi.token"]
        path = Path(self._file_path, token.path)
        if path.exists():
            return [str(token.path)]
        else:
            return []

    def get_member(self, name):
        token: Token = self.environ["manabi.token"]
        path = token.path
        if Path(name) != path:
            raise DAVError(HTTP_FORBIDDEN)
        return super().get_member(name)

    def create_empty_resource(self, name):
        raise DAVError(HTTP_FORBIDDEN)

    def create_collection(self, name):
        raise DAVError(HTTP_FORBIDDEN)

    def delete(self):
        raise DAVError(HTTP_FORBIDDEN)

    def copy_move_single(self, dest_path, is_move):
        raise DAVError(HTTP_FORBIDDEN)

    def support_recursive_move(self, dest_path):
        return False

    def move_recursive(self, dest_path):
        raise DAVError(HTTP_FORBIDDEN)

    def set_last_modified(self, dest_path, time_stamp, dry_run):
        raise DAVError(HTTP_FORBIDDEN)


class ManabiFileResource(FileResource):
    def delete(self):
        raise DAVError(HTTP_FORBIDDEN)

    def copy_move_single(self, dest_path, is_move):
        raise DAVError(HTTP_FORBIDDEN)

    def support_recursive_move(self, dest_path):
        return False

    def move_recursive(self, dest_path):
        raise DAVError(HTTP_FORBIDDEN)


class ManabiProvider(FilesystemProvider):
    def get_resource_inst(self, path: str, environ: Dict[str, Any]):
        token: Token = environ["manabi.token"]
        dir_access = environ["manabi.dir_access"]
        if dir_access:
            assert token.path
            path = f"/{str(token.path.parent)}"
        fp = self._loc_to_file_path(path, environ)
        if dir_access or Path(fp).is_dir():
            return ManabiFolderResource(path, environ, fp)
        else:
            path = token.path_as_url()
            fp = self._loc_to_file_path(path, environ)
            if Path(fp).exists():
                return ManabiFileResource(path, environ, fp)
            else:
                return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['manabi']

package_data = \
{'': ['*']}

install_requires = \
['WsgiDAV>=4.0.1,<5.0.0',
 'attrs>=21.4.0,<22.0.0',
 'pybase62>=0.4.3,<0.5.0',
 'pybranca>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'manabi',
    'version': '0.5.2',
    'description': 'Provide WebDAV access for documents.',
    'long_description': '# Manabi\n\n## Install\n\nMake sure libsodium exists on the system, for example execute:\n\n```bash\napk add --no-cache libsodium\napt-get install -y libsodium23\n```\n\n# Dev\n\nEnable dev-env:\n\n```bash\npyenv install 3.10.2\npoetry env use $HOME/.pyenv/versions/3.10.2/bin/python3.10\npoetry install\npoetry shell\n```\n\n## Config\n\nCall `manabi-keygen` and add the key to `config["manabi"]["key"]`. The key is\nshared between the caluma/alexandria backend and the WebDAV server.\n\n`mount_path`\n\nPrefix that gets passed to wsgidav, if URL rewrites remove any prefixes use `"/"`\n\n`lock_storage`\n\nThe ManabiLockLockStorage forces the WebDav log-timeout to `token-refresh-time / 2`\n\n`provider_mapping`\n\nExtends the FilesystemProvider any will only serve files if the token is valid\n\n`middleware_stack`\n\nBased on the default middleware_stack but HTTPAuthenticator is replace by\nManabiAuthenticator, which validates the tokens.\n\n`manabi.key`\n\nShared-key between the server that creates tokens to grant access and wsgi-dav\n\n`manabi.refresh`\n\nHow often tokens are refreshed in seconds, we recommend 10 minutes: `600`\n\n`manabi.initial`\n\nThe time from the token being generated till it has to be refreshed the first\ntime, we recommend 1 minues: `60`. In case tokens leak, for example via cache on\na computer, tokens should be expired by the time an adversary gets them.\n\n```python\nconfig = {\n    "mount_path": "/dav",\n    "lock_storage": ManabiLockLockStorage(refresh),\n    "provider_mapping": {\n        "/": ManabiProvider(settings.MEDIA_ROOT),\n    },\n    "middleware_stack": [\n        WsgiDavDebugFilter,\n        ErrorPrinter,\n        ManabiAuthenticator,\n        WsgiDavDirBrowser,\n        RequestResolver,\n    ],\n    "manabi": {\n        "key": key,\n        "refresh": refresh,\n        "initial": settings.MANABI_TOKEN_ACTIVATE_TIMEOUT,\n    },\n}\n```\n',
    'author': 'Adfinis AG',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/projectcaluma/manabi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



import shutil
from pathlib import Path
from subprocess import run
from typing import Any, Dict, Generator
from unittest import mock as unitmock

import pytest  # type: ignore

from . import mock
from .log import verbose_logging


@pytest.fixture(scope="session", autouse=True)
def enable_verbose_logging() -> None:
    verbose_logging()


@pytest.fixture()
def lock_storage():
    with mock.lock_storage() as storage:
        yield storage


@pytest.fixture()
def server_dir() -> Path:
    return mock.get_server_dir()


@pytest.fixture()
def config(server_dir, lock_storage) -> Dict[str, Any]:
    return mock.get_config(server_dir, lock_storage)


@pytest.fixture()
def server(config: Dict[str, Any]) -> Generator:
    with mock.run_server(config):
        yield


@pytest.fixture()
def mock_now_invalid_past() -> Generator[unitmock.MagicMock, None, None]:
    with mock.shift_now(-1200) as m:
        yield m


@pytest.fixture()
def mock_now_invalid_future() -> Generator[unitmock.MagicMock, None, None]:
    with mock.shift_now(1200) as m:
        yield m


@pytest.fixture(scope="module")
def cargo():
    if shutil.which("cargo"):
        with mock.branca_impl():
            run(["cargo", "run", "x", "y"])
